from .domain import domain_to_storename, store_domain
from .hmac import validate
from .httpclient import HTTPClient
from .jwtoken import JWTBaseModel
from .misc import get_unique_id, now_epoch

__all__ = [
    'now_epoch',
    'get_unique_id',
    'validate',
    'JWTBaseModel',
    'HTTPClient',
    'domain_to_storename',
    'store_domain',
]



from asyncio import get_event_loop
from os import environ

import nest_asyncio  # type: ignore
from pytest import fixture

nest_asyncio.apply()

environ['SHOPIFY_API_KEY'] = 'API_KEY'
environ['SHOPIFY_SECRET_KEY'] = 'SECRET_KEY'
environ['SHOPIFY_handle'] = 'HANDLE'


@fixture()
def event_loop():
    """Prevent errors where two coroutines are in different loops by never closing the loop"""
    loop = get_event_loop()
    yield loop



from asyncio import sleep
from math import floor
from time import monotonic
from typing import Any, Dict, Optional

from httpx import AsyncClient, Response
from loguru import logger
from tenacity import retry
from tenacity.retry import retry_if_exception, retry_if_exception_type
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random

from .constants import API_VERSION, MAX_TOKENS, RATE, THROTTLED_ERROR_MESSAGE
from .exceptions import (
    ShopifyCallInvalidError,
    ShopifyError,
    ShopifyThrottledError,
    not_our_fault,
)


class Store:
    _instances: dict = {}

    def __init__(self, store_id: str, name: str, access_token: str):
        self.id = store_id
        self.name = name
        self.url = f'https://{name}.myshopify.com/admin/api/{API_VERSION}'

        self.client = AsyncClient()
        self.updated_at = monotonic()

        self.reset_access_token(access_token=access_token)

    @classmethod
    def load(cls, store_id: str, name: str, access_token: Optional[str]):
        """Load the store from memory to reuse the tokens

        WARNING: the name will not be changed here after the first initialization
        """
        if store_id not in Store._instances:
            if access_token is None:
                message = 'Store {name} ({store_id}) initialized without an access_token'
                logger.error(message)
                raise ValueError(message)
            Store._instances[store_id] = Store(
                store_id=store_id, name=name, access_token=access_token
            )
        else:
            # Verify if the access token has changed
            if (
                access_token is not None
                and Store._instances[store_id].access_token != access_token
            ):
                Store._instances[store_id].reset_access_token(access_token)
        return Store._instances[store_id]

    def reset_access_token(self, access_token: str):
        """Use this function to initialize a new access token for this store"""
        self.access_token = access_token
        self.access_token_invalid = False
        self.tokens = MAX_TOKENS
        self.max_tokens = MAX_TOKENS
        self.rate = RATE

    async def __wait_for_token(self):
        self.__add_new_tokens()
        while self.tokens <= 1:
            await sleep(1)
            self.__add_new_tokens()
        self.tokens -= 1

    def __add_new_tokens(self):
        now = monotonic()
        time_since_update = now - self.updated_at
        new_tokens = floor(time_since_update * self.rate)
        if new_tokens > 1:
            self.tokens = min(self.tokens + new_tokens, self.max_tokens)
            self.updated_at = now

    async def __handle_error(self, debug: str, endpoint: str, response: Response):
        """Handle any error that occured when calling Shopify

        If the response has a valid json then return it too.
        """
        msg = (
            f'ERROR in store {self.name}: {debug}\n'
            f'API response code: {response.status_code}\n'
            f'API endpoint: {endpoint}\n'
        )
        try:
            jresp = response.json()
        except Exception:
            pass
        else:
            msg += f'API response json: {jresp}\n'

        if 400 <= response.status_code < 500:
            # This appears to be our fault
            raise ShopifyCallInvalidError(msg)

        raise ShopifyError(msg)

    @retry(
        reraise=True,
        wait=wait_random(min=1, max=2),
        stop=stop_after_attempt(5),
        retry=retry_if_exception(not_our_fault),
    )
    async def shoprequest(self, goodstatus, debug, endpoint, **kwargs) -> Dict[str, Any]:
        while True:
            await self.__wait_for_token()
            kwargs['url'] = f'{self.url}{endpoint}'
            kwargs['headers'] = {'X-Shopify-Access-Token': self.access_token}

            response = await self.client.request(**kwargs)
            if response.status_code == 429:
                # We hit the limit, we are out of tokens
                self.tokens = 0
                continue
            elif 400 <= response.status_code or response.status_code != goodstatus:
                # All errors are handled here
                await self.__handle_error(debug=debug, endpoint=endpoint, response=response)
            else:
                jresp = response.json()
                # Recalculate the rate to be sure we have the right one.
                calllimit = response.headers['X-Shopify-Shop-Api-Call-Limit']
                self.max_tokens = int(calllimit.split('/')[1])
                # In Shopify the bucket is emptied after 20 seconds
                # regardless of the bucket size.
                self.rate = int(self.max_tokens / 20)

            return jresp

    @retry(
        reraise=True,
        wait=wait_random(min=1, max=2),
        stop=stop_after_attempt(5),
        retry=retry_if_exception_type(ShopifyThrottledError),
    )
    async def execute_gql(self, query: str, variables: Dict[str, Any] = {}) -> Dict[str, Any]:
        """Simple graphql query executor because python has no decent graphql client"""

        url = f'{self.url}/graphql.json'
        headers = {
            'Content-type': 'application/json',
            'X-Shopify-Access-Token': self.access_token,
        }
        resp = await self.client.post(
            url=url, json={'query': query, 'variables': variables}, headers=headers
        )
        jsondata = resp.json()
        if type(jsondata) is not dict:
            raise ValueError('JSON data is not a dictionary')
        if 'Invalid API key or access token' in jsondata.get('errors', ''):
            self.access_token_invalid = True
            logger.warning(
                f'Store {self.name}: The Shopify API token is invalid. '
                'Flag the access token as invalid.'
            )
            raise ConnectionRefusedError
        if 'data' not in jsondata and 'errors' in jsondata:
            errorlist = '\n'.join(
                [err['message'] for err in jsondata['errors'] if 'message' in err]
            )
            if THROTTLED_ERROR_MESSAGE in errorlist:
                raise ShopifyThrottledError(
                    f'Store {self.name}: The Shopify API token is throttling. '
                )
            else:
                raise ValueError(f'GraphQL query is incorrect:\n{errorlist}')

        return jsondata['data']


class UniqueStore(Store):
    __instance = None

    def __new__(cls):
        if UniqueStore.__instance is None:
            raise ValueError('The unique store was not initialized')
        return UniqueStore.__instance

    @staticmethod
    def init(store_id, name, access_token):
        UniqueStore.__instance = Store(store_id, name, access_token)



from pydantic import BaseSettings


class Configuration(BaseSettings):
    api_key: str
    secret_key: str
    handle: str

    class Config:
        env_prefix = 'SHOPIFY_'  # defaults to no prefix, i.e. ""


conf = Configuration()



from typing import List
from unittest.mock import AsyncMock
from urllib.parse import ParseResult, parse_qs, urlencode, urlparse

import pytest
from box import Box  # type: ignore
from fastapi import FastAPI
from fastapi.testclient import TestClient
from pydantic.dataclasses import dataclass
from requests import Response

from spylib.oauth import OfflineToken, OnlineToken, conf, init_oauth_router
from spylib.utils import JWTBaseModel, hmac, now_epoch

TEST_STORE = 'test.myshopify.com'
TEST_DATA = Box(
    dict(
        app_scopes=['write_products', 'read_customers'],
        user_scopes=['write_orders', 'read_products'],
        public_domain='test.testing.com',
        private_key='TESTPRIVATEKEY',
        post_install=AsyncMock(return_value=JWTBaseModel()),
        post_login=AsyncMock(return_value=None),
    )
)

OFFLINETOKEN_DATA = dict(access_token='OFFLINETOKEN', scope=','.join(TEST_DATA.app_scopes))
ONLINETOKEN_DATA = dict(
    access_token='ONLINETOKEN',
    scope=','.join(TEST_DATA.app_scopes),
    expires_in=86399,
    associated_user_scope=','.join(TEST_DATA.user_scopes),
    associated_user={
        "id": 902541635,
        "first_name": "John",
        "last_name": "Smith",
        "email": "john@example.com",
        "email_verified": True,
        "account_owner": True,
        "locale": "en",
        "collaborator": False,
    },
)


@dataclass
class MockHTTPResponse:
    status_code: int
    jsondata: dict
    headers: dict = None  # type: ignore

    def json(self):
        return self.jsondata


@pytest.mark.asyncio
async def test_oauth(mocker):

    app = FastAPI()

    oauth_router = init_oauth_router(**TEST_DATA)

    app.include_router(oauth_router)
    client = TestClient(app)

    # --------- Test the initialization endpoint -----------

    # Missing shop argument
    response = client.get('/shopify/auth')
    assert response.status_code == 422
    assert response.json() == {
        'detail': [
            {'loc': ['query', 'shop'], 'msg': 'field required', 'type': 'value_error.missing'}
        ],
    }

    # Happy path
    response = client.get('/shopify/auth', params=dict(shop=TEST_STORE), allow_redirects=False)
    query = check_oauth_redirect_url(
        response=response, client=client, path='/admin/oauth/authorize', scope=TEST_DATA.app_scopes
    )
    state = check_oauth_redirect_query(query=query, scope=TEST_DATA.app_scopes)

    # Callback calls to get tokens
    shopify_request_mock = mocker.patch('httpx.AsyncClient.request', new_callable=AsyncMock)
    shopify_request_mock.side_effect = [
        MockHTTPResponse(status_code=200, jsondata=OFFLINETOKEN_DATA),
        MockHTTPResponse(status_code=200, jsondata=ONLINETOKEN_DATA),
    ]
    # --------- Test the callback endpoint for installation -----------
    query_str = urlencode(
        dict(shop=TEST_STORE, state=state, timestamp=now_epoch(), code='INSTALLCODE')
    )
    hmac_arg = hmac.calculate_from_message(secret=conf.secret_key, message=query_str)
    query_str += '&hmac=' + hmac_arg

    response = client.get('/callback', params=query_str, allow_redirects=False)
    query = check_oauth_redirect_url(
        response=response,
        client=client,
        path='/admin/oauth/authorize',
        scope=TEST_DATA.user_scopes,
    )
    state = check_oauth_redirect_query(
        query=query,
        scope=TEST_DATA.user_scopes,
        query_extra={'grant_options[]': ['per-user']},
    )

    assert await shopify_request_mock.called_with(
        method='post',
        url=f'https://{TEST_STORE}/admin/oauth/access_token',
        json={'client_id': conf.api_key, 'client_secret': conf.secret_key, 'code': 'INSTALLCODE'},
    )

    TEST_DATA.post_install.assert_called_once()
    TEST_DATA.post_install.assert_called_with('test', OfflineToken(**OFFLINETOKEN_DATA))

    # --------- Test the callback endpoint for login -----------
    query_str = urlencode(
        dict(shop=TEST_STORE, state=state, timestamp=now_epoch(), code='LOGINCODE'), safe='=,&/[]:'
    )
    hmac_arg = hmac.calculate_from_message(secret=conf.secret_key, message=query_str)
    query_str += '&hmac=' + hmac_arg

    response = client.get('/callback', params=query_str, allow_redirects=False)
    state = check_oauth_redirect_url(
        response=response,
        client=client,
        path=f'/admin/apps/{conf.handle}',
        scope=TEST_DATA.user_scopes,
    )

    assert await shopify_request_mock.called_with(
        method='post',
        url=f'https://{TEST_STORE}/admin/oauth/access_token',
        json={'client_id': conf.api_key, 'client_secret': conf.secret_key, 'code': 'LOGINCODE'},
    )

    TEST_DATA.post_login.assert_called_once()
    TEST_DATA.post_login.assert_called_with('test', OnlineToken(**ONLINETOKEN_DATA))


def check_oauth_redirect_url(response: Response, client, path: str, scope: List[str]) -> str:
    print(response.text)
    assert response.status_code == 307

    parsed_url = urlparse(client.get_redirect_target(response))

    expected_parsed_url = ParseResult(
        scheme='https',
        netloc=TEST_STORE,
        path=path,
        query=parsed_url.query,  # We check that separately
        params='',
        fragment='',
    )
    assert parsed_url == expected_parsed_url

    return parsed_url.query


def check_oauth_redirect_query(query: str, scope: List[str], query_extra: dict = {}) -> str:
    parsed_query = parse_qs(query)
    state = parsed_query.pop('state', [''])[0]

    expected_query = dict(
        client_id=[conf.api_key],
        redirect_uri=[f'https://{TEST_DATA.public_domain}/callback'],
        scope=[','.join(scope)],
    )
    expected_query.update(query_extra)

    assert parsed_query == expected_query

    return state



from time import monotonic
from unittest.mock import AsyncMock

import pytest
from pydantic import validator
from pydantic.dataclasses import dataclass

from spylib import Store
from spylib.exceptions import ShopifyCallInvalidError


@dataclass
class MockHTTPResponse:
    status_code: int
    jsondata: dict
    headers: dict = None  # type: ignore

    @validator('headers', pre=True, always=True)
    def set_id(cls, fld):
        return fld or {'X-Shopify-Shop-Api-Call-Limit': '39/40'}

    def json(self):
        return self.jsondata


@pytest.mark.asyncio
async def test_store_rest_happypath(mocker):
    store = Store(store_id='TEST', name='test-store', access_token='Te5tM3')

    shopify_request_mock = mocker.patch(
        'httpx.AsyncClient.request',
        new_callable=AsyncMock,
        return_value=MockHTTPResponse(status_code=200, jsondata={'success': True}),
    )

    jsondata = await store.shoprequest(
        goodstatus=200, debug='Test failed', endpoint='/test.json', method='get'
    )

    shopify_request_mock.assert_called_once()

    assert jsondata == {'success': True}

    # 80 from assuming Shopify plus then 1 used just now.
    assert store.tokens == 79


@pytest.mark.asyncio
async def test_store_rest_badrequest(mocker):
    store = Store(store_id='TEST', name='test-store', access_token='Te5tM3')

    shopify_request_mock = mocker.patch(
        'httpx.AsyncClient.request',
        new_callable=AsyncMock,
        return_value=MockHTTPResponse(
            status_code=422, jsondata={'errors': {'title': ["can't be blank"]}}
        ),
    )

    with pytest.raises(ShopifyCallInvalidError):
        await store.shoprequest(
            goodstatus=201,
            debug='Test failed',
            endpoint='/products.json',
            method='post',
            json={'product': {'body_html': 'A mystery!'}},
        )

    shopify_request_mock.assert_called_once()


params = [
    pytest.param(0, 1000, 79, id='Last call hit rate limit, long time ago'),
    pytest.param(0, 20, 79, id='Last call hit rate limit, 20s ago'),
    pytest.param(0, 10, 39, id='Last call hit rate limit, 10s ago'),
    # Wait 1 second to get 4 then use 1 so 3
    pytest.param(0, 0, 3, id='Last call that hit rate limit just happened'),
]


@pytest.mark.parametrize('init_tokens, time_passed, expected_tokens', params)
@pytest.mark.asyncio
async def test_store_rest_ratetokens(init_tokens, time_passed, expected_tokens, mocker):
    store = Store(store_id='TEST', name='test-store', access_token='Te5tM3')

    # Simulate that there is only 2 calls available before hitting the rate limit.
    # If we set this to zero, then the code will wait 1 sec which is not great to keep the tests
    # fast
    store.tokens = init_tokens
    store.updated_at = monotonic() - time_passed  # A looooong time ago

    shopify_request_mock = mocker.patch(
        'httpx.AsyncClient.request',
        new_callable=AsyncMock,
        return_value=MockHTTPResponse(status_code=200, jsondata={'success': True}),
    )
    await store.shoprequest(
        goodstatus=200, debug='Test failed', endpoint='/test.json', method='get'
    )

    shopify_request_mock.assert_called_once()

    assert store.tokens == expected_tokens


@pytest.mark.asyncio
async def test_store_graphql_happypath(mocker):
    store = Store(store_id='TEST', name='test-store', access_token='Te5tM3')

    query = '''
    {
      shop {
        name
      }
    }'''
    data = {'shop': {'name': 'graphql-admin'}}
    gql_response = {
        'data': data,
        'extensions': {
            'cost': {
                'requestedQueryCost': 1,
                'actualQueryCost': 1,
                'throttleStatus': {
                    'maximumAvailable': 1000,
                    'currentlyAvailable': 999,
                    'restoreRate': 50,
                },
            }
        },
    }

    shopify_request_mock = mocker.patch(
        'httpx.AsyncClient.request',
        new_callable=AsyncMock,
        return_value=MockHTTPResponse(status_code=200, jsondata=gql_response),
    )

    jsondata = await store.execute_gql(query=query)

    shopify_request_mock.assert_called_once()

    assert jsondata == data


@pytest.mark.asyncio
async def test_store_graphql_badquery(mocker):
    store = Store(store_id='TEST', name='test-store', access_token='Te5tM3')

    query = '''
    {
      shopp {
        name
      }
    }'''
    error_msg = "Field 'shopp' doesn't exist on type 'QueryRoot'"
    gql_response = {
        'errors': [
            {
                'message': error_msg,
                'locations': [{'line': 2, 'column': 3}],
                'path': ['query', 'shopp'],
                'extensions': {
                    'code': 'undefinedField',
                    'typeName': 'QueryRoot',
                    'fieldName': 'shopp',
                },
            }
        ]
    }

    shopify_request_mock = mocker.patch(
        'httpx.AsyncClient.request',
        new_callable=AsyncMock,
        return_value=MockHTTPResponse(status_code=200, jsondata=gql_response),
    )

    with pytest.raises(ValueError, match=f'^GraphQL query is incorrect:\n{error_msg}$'):
        await store.execute_gql(query=query)

    shopify_request_mock.assert_called_once()


@pytest.mark.asyncio
async def test_store_graphql_tokeninvalid(mocker):
    store = Store(store_id='TEST', name='test-store', access_token='INVALID')

    query = '''
    {
      shop {
        name
      }
    }'''
    gql_response = {
        'errors': '[API] Invalid API key or access token (unrecognized login or wrong password)'
    }

    shopify_request_mock = mocker.patch(
        'httpx.AsyncClient.request',
        new_callable=AsyncMock,
        return_value=MockHTTPResponse(status_code=200, jsondata=gql_response),
    )

    with pytest.raises(ConnectionRefusedError):
        await store.execute_gql(query=query)

    shopify_request_mock.assert_called_once()


@pytest.mark.asyncio
async def test_store_graphql_throttling(mocker):
    store = Store(store_id='TEST', name='test-store', access_token='Te5tM3')

    query = '''
    {
      shop {
        name
      }
    }'''
    extensions = {
        "cost": {
            "requestedQueryCost": 752,
            "actualQueryCost": None,
            "throttleStatus": {
                "maximumAvailable": 1000,
                "currentlyAvailable": 662,
                "restoreRate": 50,
            },
        }
    }

    gql_response_throttled = {
        "errors": [
            {
                "message": "Throttled",
                "extensions": {
                    "code": "THROTTLED",
                    "documentation": "https://help.shopify.com/api/graphql-admin-api/"
                    "graphql-admin-api-rate-limits",
                },
            }
        ],
        "extensions": extensions,
    }
    data = {'shop': {'name': 'graphql-admin'}}
    gql_response = {'data': data, 'extensions': extensions}

    shopify_request_mock = mocker.patch(
        'httpx.AsyncClient.request',
        new_callable=AsyncMock,
        side_effect=[
            MockHTTPResponse(status_code=200, jsondata=gql_response_throttled),
            MockHTTPResponse(status_code=200, jsondata=gql_response),
        ],
    )

    assert 'shop' in await store.execute_gql(query=query)
    shopify_request_mock.call_count == 2



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['spylib', 'spylib.oauth', 'spylib.utils']

package_data = \
{'': ['*']}

setup(name='spylib',
      version='0.2.2',
      description="A library to facilitate interfacing with Shopify's API",
      author='Anthony Hillairet',
      author_email='ant@satel.ca',
      url='https://github.com/SatelCreative/satel-spylib',
      packages=packages,
      package_data=package_data,
     )
